@@ -1,26 +1,39 @@
 # Standard lib imports
 import random
+import logging
+import typing
 
 # Local imports
 from qwilfish.constants import DEFAULT_START_SYMBOL
-from qwilfish.constants import DEFAULT_MIN_NONTERMINALS
-from qwilfish.constants import DEFAULT_MAX_NONTERMINALS
 from qwilfish.grammar import GrammarError
 from qwilfish.grammar import validate
 from qwilfish.grammar import nonterminals
 from qwilfish.grammar import is_nonterminal
 from qwilfish.grammar import expansion_opt
 from qwilfish.grammar import expansion_string
+from qwilfish.grammar import EXP_OPTS_INDEX
 from qwilfish.derivation_tree import DerivationTreeError
 from qwilfish.derivation_tree import expansion_to_children
 from qwilfish.derivation_tree import all_terminals
 
+log = logging.getLogger(__name__)
+
 class QwilFuzzer():
 
     OPT_PRE = "pre"
     OPT_POST = "post"
     OPT_PROB = "prob"
-    DEFAULT_OPTS = [OPT_PRE, OPT_POST, OPT_PROB]
+    OPT_PROBFB = "probfb"
+    DEFAULT_OPTS = [OPT_PRE, OPT_POST, OPT_PROB, OPT_PROBFB]
+    DEFAULT_MIN_NONTERMINALS = 0
+    DEFAULT_MAX_NONTERMINALS = 1000
+
+    # Default probability increase when updating probf annotations
+    FB_FACTOR = 0.1
+    # Tolerance when summing probabilities
+    PROB_EPSILON = 0.00001
+    # Minimum probability for entries being decreased
+    PROB_MIN = 0.001
 
     def __init__(self,
                  grammar,
@@ -28,18 +41,18 @@ class QwilFuzzer():
                  min_nonterminals=DEFAULT_MIN_NONTERMINALS,
                  max_nonterminals=DEFAULT_MAX_NONTERMINALS,
                  supported_opts=DEFAULT_OPTS,
-                 debug=False):
+                 unguided_fuzzing=False):
         self.grammar = grammar
         self.start_symbol = start_symbol
         self.min_nonterminals = min_nonterminals
         self.max_nonterminals = max_nonterminals
+        self.unguided_fuzzing = unguided_fuzzing
 
         self.supported_opts = set()
         if supported_opts:
             for i in range(0, len(supported_opts)):
                 self.supported_opts |= {supported_opts[i]}
 
-        self.debug=debug
         self.check_grammar()
 
     def fuzz(self):
@@ -59,6 +72,115 @@ class QwilFuzzer():
 
         return True
 
+    def get_probabilities(self):
+        probs = {}
+
+        for nonterminal in self.grammar:
+            for expansion in self.grammar[nonterminal]:
+                prob = self.opt_prob_all(expansion) # include fixed and dynamic
+                if prob:
+                    probs.update(
+                        {f"{nonterminal}.{expansion_string(expansion)}": prob})
+
+        return probs
+
+    def update_grammar_probabilities(self, subtree):
+        # Updating the grammar has been disabled for this instance, return
+        if self.unguided_fuzzing:
+            return
+
+        (symbol, children) = subtree
+
+        # Terminal symbol
+        if not children:
+            return
+
+        # Attempt to recreate the expansion string by joining the symbols
+        # from all the first-level children
+        wanted_expansion = "".join([c[0] for c in children])
+
+        # Possible expansions for this symbol
+        expansions = self.grammar[symbol]
+        chosen_exp = None
+        chosen_idx = None # Index of the chosen expansion in this subtree
+        dynamic_idx = [] # Index of all dynamic (probf) expansions
+        min_idx = []# Index of all dynamic expansions with a minimized probfb
+        sum_static = 0
+
+        for i,e in enumerate(expansions):
+            if self.opt_prob(e):
+                sum_static += self.opt_prob(e)
+            elif self.opt_probfb(e):
+                dynamic_idx.append(i)
+                if wanted_expansion == expansion_string(e):
+                    chosen_exp = e
+                    chosen_idx = i
+                if abs(self.opt_probfb(e) - QwilFuzzer.PROB_MIN) < \
+                   QwilFuzzer.PROB_EPSILON:
+                   min_idx.append(i)
+
+        if chosen_exp:
+            # Maximum probability that a dynamic (probfb) annotation can have.
+            # Formula is:
+            # 1.0 - ([static prob tot] + [N probf annots]*[min dynamic prob])
+            max_dyn = 1.0 - (sum_static + len(dynamic_idx)*QwilFuzzer.PROB_MIN)
+
+            # Calculate how much we can add to the chosen expansion from each
+            # unused and non-minimized dynamic expansions without breaking the
+            # max limit
+            maxlim_diff = float("inf")
+            if self.opt_probfb(chosen_exp) < max_dyn:
+                maxlim_diff = max_dyn - self.opt_probfb(chosen_exp)
+                # Calculate how much all unused dynamic probs will be lowered
+                maxlim_diff /= len(dynamic_idx) - len(min_idx)
+            #else:
+            #    chosen_exp[EXP_OPTS_INDEX][QwilFuzzer.OPT_PROBFB] = max_dyn
+
+            # Calculate the default value to add to the chosen expansion from
+            # each unused and non-minimized dynamic expansions.
+            def_diff = QwilFuzzer.FB_FACTOR/(len(dynamic_idx) - len(min_idx))
+
+
+            # Calculate how much we can add to the chosen expansion from each
+            # unused and non-minimized dynamic expansions without causing
+            # any of them to fall below the min limit.
+            minlim_diff = float("inf")
+            for i in dynamic_idx:
+                p = self.opt_probfb(expansions[i])
+                if abs(p - QwilFuzzer.PROB_MIN) < QwilFuzzer.PROB_EPSILON:
+                    continue
+                    #expansions[i][EXP_OPTS_INDEX][QwilFuzzer.OPT_PROBFB] = \
+                    #    QwilFuzzer.PROB_MIN
+                elif p > QwilFuzzer.PROB_MIN:
+                    if minlim_diff > (p - QwilFuzzer.PROB_MIN):
+                        minlim_diff = p - QwilFuzzer.PROB_MIN
+                else: # Should not happen
+                    log.warning(f"Unexpected probfb value for {expansions[i]}")
+                    expansions[i][EXP_OPTS_INDEX][QwilFuzzer.OPT_PROBFB] = \
+                        QwilFuzzer.PROB_MIN
+
+            diff = min(minlim_diff, def_diff, maxlim_diff)
+
+            log.info(f"{symbol} before: {expansions}")
+
+            for i in dynamic_idx:
+                # Redistribute "diff" amount of probability from every
+                # non-minimized and unused dynamic (probfb) expansion to
+                # the expansion that was used in this particular subtree
+                if i in min_idx:
+                    continue
+                elif i == chosen_idx:
+                    continue
+                else:
+                    e = expansions[i]
+                    e[EXP_OPTS_INDEX][QwilFuzzer.OPT_PROBFB] -= diff
+                    chosen_exp[EXP_OPTS_INDEX][QwilFuzzer.OPT_PROBFB] += diff
+
+            log.info(f"{symbol} after: {expansions}")
+
+        for c in children:
+            self.update_grammar_probabilities(c)
+
 ###############################################################################
 ############################################################### Private methods
 ###############################################################################
@@ -73,14 +195,12 @@ class QwilFuzzer():
         return expansion_to_children(expansion)
 
     def expand_node_min_cost(self, node):
-        if self.debug:
-            print("Expanding", all_terminals(node), "at minimum cost")
+        log.debug("Expanding %s at minimum cost", all_terminals(node))
 
         return self.expand_node_by_cost(node, min)
 
     def expand_node_max_cost(self, node):
-        if self.debug:
-            print("Expanding", all_terminals(node), "at maximum cost")
+        log.debug("Expanding %s at maximum cost", all_terminals(node))
 
         return self.expand_node_by_cost(node, max)
 
@@ -90,8 +210,7 @@ class QwilFuzzer():
             raise DerivationTreeError("Unexpandable node, children: " +
                                       children)
 
-        if self.debug:
-            print("Expanding", all_terminals(node), "randomly")
+        log.debug("Expanding %s randomly", all_terminals(node))
 
         expansions = self.grammar[symbol]
         children_alternatives = \
@@ -114,6 +233,19 @@ class QwilFuzzer():
     def opt_prob(self, expansion):
         return expansion_opt(expansion, QwilFuzzer.OPT_PROB)
 
+    def opt_prob_all(self, expansion):
+        prob = expansion_opt(expansion, QwilFuzzer.OPT_PROB)
+
+        # If no regular (static) prob opt has been defined, look for a
+        # feedback (dynamic) prob opt.
+        if not prob:
+            prob = expansion_opt(expansion, QwilFuzzer.OPT_PROBFB)
+
+        return prob
+
+    def opt_probfb(self, expansion):
+        return expansion_opt(expansion, QwilFuzzer.OPT_PROBFB)
+
     def process_chosen_children(self, children, expansion):
         function = self.opt_pre_func(expansion)
         if function is None:
@@ -124,8 +256,7 @@ class QwilFuzzer():
 
         result = function()
 
-        if self.debug:
-            print(repr(function) + "()", "=", repr(result))
+        log.debug("%s()=%s", repr(function), repr(result))
 
         return self.apply_result(result, children)
 
@@ -141,9 +272,8 @@ class QwilFuzzer():
                     child_index = symbol_indexes[index]
                     if not isinstance(value, str):
                         value = repr(value)
-                    if self.debug:
-                        print("Replacing", all_terminals(
-                            children[child_index]), "by", value)
+                    log.debug("Replacing %s by %s",
+                              all_terminals(children[child_index]), value)
 
                     child_symbol, _ = children[child_index]
                     children[child_index] = (child_symbol, [(value, [])])
@@ -154,9 +284,8 @@ class QwilFuzzer():
         elif isinstance(result, bool):
             pass
         else:
-            if self.debug:
-                print("Replacing", "".join(
-                    [all_terminals(c) for c in children]), "by", result)
+            log.debug("Replacing %s by %s",
+                "".join([all_terminals(c) for c in children]), result)
 
             children = [(repr(result), [])]
 
@@ -171,8 +300,7 @@ class QwilFuzzer():
         for children in children_alternatives:
             expansion = all_terminals((symbol, children))
             children_weight = probabilities[expansion]
-            if self.debug:
-                print(repr(expansion), "p =", children_weight)
+            log.debug("p(%s) = %s", repr(expansion), children_weight)
             weights.append(children_weight)
 
         if sum(weights) == 0:
@@ -246,9 +374,8 @@ class QwilFuzzer():
         if function is not None:
             result = self.eval_function(tree, function)
             if isinstance(result, bool) and not result:
-                if self.debug:
-                    print(all_terminals(tree),
-                        "did not satisfy", symbol, "constraint")
+                log.debug("%s did not satisfy %s constraint",
+                          all_terminals(tree), symbol)
                 return False, children
 
             children = self.apply_result(result, children)
@@ -293,8 +420,7 @@ class QwilFuzzer():
                 args.append(symbol_value)
 
         result = function(tree, *args)
-        if self.debug:
-            print(repr(function) + repr(tuple(args)), "=", repr(result))
+        log.debug("%s%s=%s", repr(function), repr(tuple(args)), repr(result))
 
         return result
 
@@ -320,7 +446,7 @@ class QwilFuzzer():
                 or self.possible_expansions(tree) < limit)
                 and self.any_possible_expansions(tree)):
             tree = self.expand_tree_once(tree)
-            self.debug_tree(tree)
+            log.debug("Tree: %s", all_terminals(tree))
 
         return tree
 
@@ -374,12 +500,9 @@ class QwilFuzzer():
 
         return sum(self.symbol_cost(s, seen) for s in symbols) + 1
 
-    def debug_tree(self, tree):
-        if self.debug:
-            print("Tree:", all_terminals(tree))
-
     def expansion_probabilities(self, expansions, nonterminal="<symbol>"):
-        probabilities = [self.opt_prob(expansion) for expansion in expansions]
+        probabilities = [self.opt_prob_all(expansion) \
+                         for expansion in expansions]
         prob_dist = self.prob_distribution(probabilities, nonterminal)
 
         prob_mapping = {}
@@ -390,12 +513,11 @@ class QwilFuzzer():
         return prob_mapping
 
     def prob_distribution(self, probabilities, nonterminal="<symbol>"):
-        epsilon = 0.00001
 
         n_unspec_probs = probabilities.count(None)
         if n_unspec_probs == 0:
-            sum_probs = cast(float, sum(probabilities))
-            if abs(sum_probs - 1.0) >= epsilon:
+            sum_probs = typing.cast(float, sum(probabilities))
+            if abs(sum_probs - 1.0) >= QwilFuzzer.PROB_EPSILON:
                 raise GrammarError(nonterminal + \
                                    ": sum of probabilities must be 1.0")
             return probabilities
@@ -417,7 +539,7 @@ class QwilFuzzer():
                 p = default_prob
             all_probs.append(p)
 
-        if abs(sum(all_probs) - 1.0) >= epsilon:
+        if abs(sum(all_probs) - 1.0) >= QwilFuzzer.PROB_EPSILON:
             raise GrammarError(nonterminal + \
                                ": couldn't set default probbilities")
 
