@@ -1,3 +1,6 @@
+# Standard lib imports
+import logging
+
 # Third-party imports
 import grpc
 
@@ -10,6 +13,8 @@ from qwilfish.generated.feedback_interface_pb2_grpc import (
     FeedbackInterfaceStub
 )
 
+log = logging.getLogger(__name__)
+
 def start_simple_client(message_type, address, port, **kwargs):
     with grpc.insecure_channel(address + ":" + str(port)) as channel:
         stub = FeedbackInterfaceStub(channel)
@@ -32,4 +37,4 @@ def start_simple_client(message_type, address, port, **kwargs):
             raise ValueError("Unrecognized message type '" + message_type + \
                              "' for qwilfish gRPC feedback interface!")
 
-    print(process_data)
+    log.debug(str(process_data))
